@@ -42,6 +42,7 @@
 #include "debug.h"
 #include "hash.h"
 #include "pkcs11.h"
+#include "pkcs11i.h"
 #include "pkcs11x.h"
 
 #include <assert.h>
@@ -100,12 +101,15 @@ attrs_build (CK_ATTRIBUTE *attrs,
 	CK_ULONG at;
 	CK_ULONG j;
 	CK_ULONG i;
+	size_t length;
 
 	/* How many attributes we already have */
 	current = p11_attrs_count (attrs);
 
 	/* Reallocate for how many we need */
-	attrs = realloc (attrs, (current + count_to_add + 1) * sizeof (CK_ATTRIBUTE));
+	length = current + count_to_add;
+	return_val_if_fail (current <= length && length < SIZE_MAX, NULL);
+	attrs = reallocarray (attrs, length + 1, sizeof (CK_ATTRIBUTE));
 	return_val_if_fail (attrs != NULL, NULL);
 
 	at = current;
@@ -669,7 +673,7 @@ attribute_is_sensitive (const CK_ATTRIBUTE *attr,
 	X (CKA_X_PEER)
 	X (CKA_X_DISTRUSTED)
 	X (CKA_X_CRITICAL)
-	X (CKA_X_PUBLIC_KEY_INFO)
+	X (CKA_PUBLIC_KEY_INFO)
 	X (CKA_NSS_URL)
 	X (CKA_NSS_EMAIL)
 	X (CKA_NSS_SMIME_INFO)
