@@ -287,15 +287,17 @@ proxy_create (Proxy **res)
 
 			return_val_if_fail (count == 0 || slots != NULL, CKR_GENERAL_ERROR);
 
-			py->mappings = realloc (py->mappings, sizeof (Mapping) * (py->n_mappings + count));
-			return_val_if_fail (py->mappings != NULL, CKR_HOST_MEMORY);
+			if (count > 0) {
+				py->mappings = realloc (py->mappings, sizeof (Mapping) * (py->n_mappings + count));
+				return_val_if_fail (py->mappings != NULL, CKR_HOST_MEMORY);
 
-			/* And now add a mapping for each of those slots */
-			for (i = 0; i < count; ++i) {
-				py->mappings[py->n_mappings].funcs = funcs;
-				py->mappings[py->n_mappings].wrap_slot = py->n_mappings + MAPPING_OFFSET;
-				py->mappings[py->n_mappings].real_slot = slots[i];
-				++py->n_mappings;
+				/* And now add a mapping for each of those slots */
+				for (i = 0; i < count; ++i) {
+					py->mappings[py->n_mappings].funcs = funcs;
+					py->mappings[py->n_mappings].wrap_slot = py->n_mappings + MAPPING_OFFSET;
+					py->mappings[py->n_mappings].real_slot = slots[i];
+					++py->n_mappings;
+				}
 			}
 
 			free (slots);
@@ -2365,7 +2367,7 @@ C_GetFunctionList (CK_FUNCTION_LIST_PTR_PTR list)
 		}
 	}
 
-	if (rv == CKR_OK && p11_virtual_can_wrap ()) {
+	if (rv == CKR_OK) {
 		state = calloc (1, sizeof (State));
 		if (!state) {
 			rv = CKR_HOST_MEMORY;
