@@ -34,6 +34,7 @@
 
 #include "config.h"
 
+#include "array.h"
 #include "attrs.h"
 #include "buffer.h"
 #define P11_DEBUG_FLAG P11_DEBUG_URI
@@ -139,15 +140,23 @@
  * Length of %P11_KIT_URI_SCHEME.
  */
 
+typedef struct _Attribute {
+	char *name;
+	char *value;
+} Attribute;
+
 struct p11_kit_uri {
 	bool unrecognized;
 	CK_INFO module;
 	CK_SLOT_INFO slot;
 	CK_TOKEN_INFO token;
 	CK_ATTRIBUTE *attrs;
+	CK_SLOT_ID slot_id;
 	char *pin_source;
 	char *pin_value;
-	CK_SLOT_ID slot_id;
+	char *module_name;
+	char *module_path;
+	p11_array *qattrs;
 };
 
 static char *
@@ -727,6 +736,173 @@ p11_kit_uri_set_pinfile (P11KitUri *uri, const char *pinfile)
 	p11_kit_uri_set_pin_source (uri, pinfile);
 }
 
+
+/**
+ * p11_kit_uri_get_module_name:
+ * @uri: The URI
+ *
+ * Get the 'module-name' part of the URI. This is used by some
+ * applications to explicitly specify the name of a PKCS\#11 module.
+ *
+ * Returns: The module-name or %NULL if not present.
+ */
+const char*
+p11_kit_uri_get_module_name (P11KitUri *uri)
+{
+	return_val_if_fail (uri != NULL, NULL);
+	return uri->module_name;
+}
+
+/**
+ * p11_kit_uri_set_module_name:
+ * @uri: The URI
+ * @name: The new module-name
+ *
+ * Set the 'module-name' part of the URI. This is used by some
+ * applications to explicitly specify the name of a PKCS\#11 module.
+ */
+void
+p11_kit_uri_set_module_name (P11KitUri *uri, const char *name)
+{
+	return_if_fail (uri != NULL);
+	free (uri->module_name);
+	uri->module_name = name ? strdup (name) : NULL;
+}
+
+/**
+ * p11_kit_uri_get_module_path:
+ * @uri: The URI
+ *
+ * Get the 'module-path' part of the URI. This is used by some
+ * applications to explicitly specify the path of a PKCS\#11 module.
+ *
+ * Returns: The module-path or %NULL if not present.
+ */
+const char*
+p11_kit_uri_get_module_path (P11KitUri *uri)
+{
+	return_val_if_fail (uri != NULL, NULL);
+	return uri->module_path;
+}
+
+/**
+ * p11_kit_uri_set_module_path:
+ * @uri: The URI
+ * @path: The new module-path
+ *
+ * Set the 'module-path' part of the URI. This is used by some
+ * applications to explicitly specify the path of a PKCS\#11 module.
+ */
+void
+p11_kit_uri_set_module_path (P11KitUri *uri, const char *path)
+{
+	return_if_fail (uri != NULL);
+	free (uri->module_path);
+	uri->module_path = path ? strdup (path) : NULL;
+}
+
+/**
+ * p11_kit_uri_get_vendor_query:
+ * @uri: The URI
+ * @name: The name of vendor query
+ *
+ * Get the vendor query part of the URI, identified by @name. This is
+ * used by some applications to explicitly specify the path of a
+ * PKCS\#11 module.
+ *
+ * Returns: The value of vendor query or %NULL if not present.
+ */
+const char*
+p11_kit_uri_get_vendor_query (P11KitUri *uri, const char *name)
+{
+	size_t i;
+
+	return_val_if_fail (uri != NULL, NULL);
+
+	for (i = 0; i < uri->qattrs->num; i++) {
+		Attribute *attr = uri->qattrs->elem[i];
+		if (strcmp (attr->name, name) == 0)
+			return attr->value;
+	}
+	return NULL;
+}
+
+static void
+free_attribute (Attribute *attr)
+{
+	free (attr->name);
+	free (attr->value);
+	free (attr);
+}
+
+static bool
+insert_attribute (p11_array *attrs, char *name, char *value)
+{
+	Attribute *attr;
+	size_t i;
+
+	return_val_if_fail (attrs != NULL, false);
+	return_val_if_fail (name != NULL, false);
+	return_val_if_fail (value != NULL, false);
+
+	for (i = 0; i < attrs->num; i++) {
+		attr = attrs->elem[i];
+		if (strcmp (attr->name, (char *)name) > 0)
+			break;
+	}
+
+	attr = calloc (1, sizeof (Attribute));
+	return_val_if_fail (attr, false);
+
+	attr->name = name;
+	attr->value = value;
+
+	return p11_array_insert (attrs, i, attr);
+}
+
+/**
+ * p11_kit_uri_set_vendor_query:
+ * @uri: The URI
+ * @name: The name of vendor query
+ * @value: (allow-none): The value of vendor query
+ *
+ * Set the vendor query part of the URI, identified by @name. This is
+ * used by some applications to explicitly specify the path of a
+ * PKCS\#11 module.
+ *
+ * Returns: 1 if the vendor query is set or removed, 0 if not.
+ */
+int
+p11_kit_uri_set_vendor_query (P11KitUri *uri, const char *name,
+			      const char *value)
+{
+	Attribute *attr;
+	size_t i;
+
+	return_val_if_fail (uri != NULL, 0);
+	return_val_if_fail (name != NULL, 0);
+
+	for (i = 0; i < uri->qattrs->num; i++) {
+		attr = uri->qattrs->elem[i];
+		if (strcmp (attr->name, name) == 0)
+			break;
+	}
+	if (i == uri->qattrs->num) {
+		if (value == NULL)
+			return 0;
+		return insert_attribute (uri->qattrs,
+					 strdup (name), strdup (value));
+	}
+	if (value == NULL)
+		p11_array_remove (uri->qattrs, i);
+	else {
+		free (attr->value);
+		attr->value = strdup (value);
+	}
+
+	return 1;
+}
+
 /**
  * p11_kit_uri_new:
  *
@@ -749,25 +925,39 @@ p11_kit_uri_new (void)
 	uri->module.libraryVersion.major = (CK_BYTE)-1;
 	uri->module.libraryVersion.minor = (CK_BYTE)-1;
 	uri->slot_id = (CK_SLOT_ID)-1;
+	uri->qattrs = p11_array_new ((p11_destroyer)free_attribute);
 
 	return uri;
 }
 
+enum uri_sep {
+	sep_path = '\0',
+	sep_pattr = ';',
+	sep_query = '?',
+	sep_qattr = '&',
+};
+
 static void
 format_name_equals (p11_buffer *buffer,
-                    bool *is_first,
+                    enum uri_sep *sep,
                     const char *name)
 {
-	if (!*is_first)
-		p11_buffer_add (buffer, ";", 1);
+	if (*sep) {
+		char c = *sep;
+		p11_buffer_add (buffer, &c, 1);
+	}
 	p11_buffer_add (buffer, name, -1);
 	p11_buffer_add (buffer, "=", 1);
-	*is_first = false;
+
+	if (*sep == sep_path)
+		*sep = sep_pattr;
+	else if (*sep == sep_query)
+		*sep = sep_qattr;
 }
 
 static bool
 format_raw_string (p11_buffer *buffer,
-                   bool *is_first,
+                   enum uri_sep *sep,
                    const char *name,
                    const char *value)
 {
@@ -775,7 +965,7 @@ format_raw_string (p11_buffer *buffer,
 	if (!value)
 		return true;
 
-	format_name_equals (buffer, is_first, name);
+	format_name_equals (buffer, sep, name);
 	p11_buffer_add (buffer, value, -1);
 
 	return p11_buffer_ok (buffer);
@@ -783,7 +973,7 @@ format_raw_string (p11_buffer *buffer,
 
 static bool
 format_encode_string (p11_buffer *buffer,
-                      bool *is_first,
+                      enum uri_sep *sep,
                       const char *name,
                       const unsigned char *value,
                       size_t n_value,
@@ -793,7 +983,7 @@ format_encode_string (p11_buffer *buffer,
 	if (!value)
 		return true;
 
-	format_name_equals (buffer, is_first, name);
+	format_name_equals (buffer, sep, name);
 	p11_url_encode (value, value + n_value, force ? "" : P11_URL_VERBATIM, buffer);
 
 	return p11_buffer_ok (buffer);
@@ -802,7 +992,7 @@ format_encode_string (p11_buffer *buffer,
 
 static bool
 format_struct_string (p11_buffer *buffer,
-                      bool *is_first,
+                      enum uri_sep *sep,
                       const char *name,
                       const unsigned char *value,
                       size_t value_max)
@@ -814,12 +1004,12 @@ format_struct_string (p11_buffer *buffer,
 		return true;
 
 	len = p11_kit_space_strlen (value, value_max);
-	return format_encode_string (buffer, is_first, name, value, len, false);
+	return format_encode_string (buffer, sep, name, value, len, false);
 }
 
 static bool
 format_attribute_string (p11_buffer *buffer,
-                         bool *is_first,
+                         enum uri_sep *sep,
                          const char *name,
                          CK_ATTRIBUTE_PTR attr,
                          bool force)
@@ -828,14 +1018,14 @@ format_attribute_string (p11_buffer *buffer,
 	if (attr == NULL)
 		return true;
 
-	return format_encode_string (buffer, is_first, name,
+	return format_encode_string (buffer, sep, name,
 	                             attr->pValue, attr->ulValueLen,
 	                             force);
 }
 
 static bool
 format_attribute_class (p11_buffer *buffer,
-                        bool *is_first,
+                        enum uri_sep *sep,
                         const char *name,
                         CK_ATTRIBUTE_PTR attr)
 {
@@ -867,12 +1057,12 @@ format_attribute_class (p11_buffer *buffer,
 		return true;
 	}
 
-	return format_raw_string (buffer, is_first, name, value);
+	return format_raw_string (buffer, sep, name, value);
 }
 
 static bool
 format_struct_version (p11_buffer *buffer,
-                       bool *is_first,
+                       enum uri_sep *sep,
                        const char *name,
                        CK_VERSION_PTR version)
 {
@@ -884,12 +1074,12 @@ format_struct_version (p11_buffer *buffer,
 
 	snprintf (buf, sizeof (buf), "%d.%d",
 	          (int)version->major, (int)version->minor);
-	return format_raw_string (buffer, is_first, name, buf);
+	return format_raw_string (buffer, sep, name, buf);
 }
 
 static bool
 format_ulong (p11_buffer *buffer,
-	      bool *is_first,
+	      enum uri_sep *sep,
 	      const char *name,
 	      CK_ULONG value)
 {
@@ -900,7 +1090,7 @@ format_ulong (p11_buffer *buffer,
 		return true;
 
 	snprintf (buf, sizeof (buf), "%lu", value);
-	return format_raw_string (buffer, is_first, name, buf);
+	return format_raw_string (buffer, sep, name, buf);
 }
 
 /**
@@ -935,7 +1125,8 @@ int
 p11_kit_uri_format (P11KitUri *uri, P11KitUriType uri_type, char **string)
 {
 	p11_buffer buffer;
-	bool is_first = true;
+	enum uri_sep sep = sep_path;
+	size_t i;
 
 	return_val_if_fail (uri != NULL, P11_KIT_URI_UNEXPECTED);
 	return_val_if_fail (string != NULL, P11_KIT_URI_UNEXPECTED);
@@ -947,10 +1138,10 @@ p11_kit_uri_format (P11KitUri *uri, P11KitUriType uri_type, char **string)
 	p11_buffer_add (&buffer, ":", 1);
 
 	if ((uri_type & P11_KIT_URI_FOR_MODULE) == P11_KIT_URI_FOR_MODULE) {
-		if (!format_struct_string (&buffer, &is_first, "library-description",
+		if (!format_struct_string (&buffer, &sep, "library-description",
 		                           uri->module.libraryDescription,
 		                           sizeof (uri->module.libraryDescription)) ||
-		    !format_struct_string (&buffer, &is_first, "library-manufacturer",
+		    !format_struct_string (&buffer, &sep, "library-manufacturer",
 		                           uri->module.manufacturerID,
 		                           sizeof (uri->module.manufacturerID))) {
 			return_val_if_reached (P11_KIT_URI_UNEXPECTED);
@@ -958,36 +1149,36 @@ p11_kit_uri_format (P11KitUri *uri, P11KitUriType uri_type, char **string)
 	}
 
 	if ((uri_type & P11_KIT_URI_FOR_MODULE_WITH_VERSION) == P11_KIT_URI_FOR_MODULE_WITH_VERSION) {
-		if (!format_struct_version (&buffer, &is_first, "library-version",
+		if (!format_struct_version (&buffer, &sep, "library-version",
 		                            &uri->module.libraryVersion)) {
 			return_val_if_reached (P11_KIT_URI_UNEXPECTED);
 		}
 	}
 
 	if ((uri_type & P11_KIT_URI_FOR_SLOT) == P11_KIT_URI_FOR_SLOT) {
-		if (!format_struct_string (&buffer, &is_first, "slot-description",
+		if (!format_struct_string (&buffer, &sep, "slot-description",
 		                           uri->slot.slotDescription,
 		                           sizeof (uri->slot.slotDescription)) ||
-		    !format_struct_string (&buffer, &is_first, "slot-manufacturer",
+		    !format_struct_string (&buffer, &sep, "slot-manufacturer",
 		                           uri->slot.manufacturerID,
 		                           sizeof (uri->slot.manufacturerID)) ||
-		    !format_ulong (&buffer, &is_first, "slot-id",
+		    !format_ulong (&buffer, &sep, "slot-id",
 				   uri->slot_id)) {
 			return_val_if_reached (P11_KIT_URI_UNEXPECTED);
 		}
 	}
 
 	if ((uri_type & P11_KIT_URI_FOR_TOKEN) == P11_KIT_URI_FOR_TOKEN) {
-		if (!format_struct_string (&buffer, &is_first, "model",
+		if (!format_struct_string (&buffer, &sep, "model",
 		                           uri->token.model,
 		                           sizeof (uri->token.model)) ||
-		    !format_struct_string (&buffer, &is_first, "manufacturer",
+		    !format_struct_string (&buffer, &sep, "manufacturer",
 		                           uri->token.manufacturerID,
 		                           sizeof (uri->token.manufacturerID)) ||
-		    !format_struct_string (&buffer, &is_first, "serial",
+		    !format_struct_string (&buffer, &sep, "serial",
 		                           uri->token.serialNumber,
 		                           sizeof (uri->token.serialNumber)) ||
-		    !format_struct_string (&buffer, &is_first, "token",
+		    !format_struct_string (&buffer, &sep, "token",
 		                           uri->token.label,
 		                           sizeof (uri->token.label))) {
 			return_val_if_reached (P11_KIT_URI_UNEXPECTED);
@@ -995,23 +1186,25 @@ p11_kit_uri_format (P11KitUri *uri, P11KitUriType uri_type, char **string)
 	}
 
 	if ((uri_type & P11_KIT_URI_FOR_OBJECT) == P11_KIT_URI_FOR_OBJECT) {
-		if (!format_attribute_string (&buffer, &is_first, "id",
+		if (!format_attribute_string (&buffer, &sep, "id",
 		                              p11_kit_uri_get_attribute (uri, CKA_ID),
 		                              true) ||
-		    !format_attribute_string (&buffer, &is_first, "object",
+		    !format_attribute_string (&buffer, &sep, "object",
 		                              p11_kit_uri_get_attribute (uri, CKA_LABEL),
 		                              false)) {
 			return_val_if_reached (P11_KIT_URI_UNEXPECTED);
 		}
 
-		if (!format_attribute_class (&buffer, &is_first, "type",
+		if (!format_attribute_class (&buffer, &sep, "type",
 		                             p11_kit_uri_get_attribute (uri, CKA_CLASS))) {
 			return_val_if_reached (P11_KIT_URI_UNEXPECTED);
 		}
 	}
 
+	sep = sep_query;
+
 	if (uri->pin_source) {
-		if (!format_encode_string (&buffer, &is_first, "pin-source",
+		if (!format_encode_string (&buffer, &sep, "pin-source",
 		                           (const unsigned char*)uri->pin_source,
 		                           strlen (uri->pin_source), 0)) {
 			return_val_if_reached (P11_KIT_URI_UNEXPECTED);
@@ -1019,13 +1212,38 @@ p11_kit_uri_format (P11KitUri *uri, P11KitUriType uri_type, char **string)
 	}
 
 	if (uri->pin_value) {
-		if (!format_encode_string (&buffer, &is_first, "pin-value",
+		if (!format_encode_string (&buffer, &sep, "pin-value",
 		                           (const unsigned char*)uri->pin_value,
 		                           strlen (uri->pin_value), 0)) {
 			return_val_if_reached (P11_KIT_URI_UNEXPECTED);
 		}
 	}
 
+	if (uri->module_name) {
+		if (!format_encode_string (&buffer, &sep, "module-name",
+		                           (const unsigned char*)uri->module_name,
+		                           strlen (uri->module_name), 0)) {
+			return_val_if_reached (P11_KIT_URI_UNEXPECTED);
+		}
+	}
+
+	if (uri->module_path) {
+		if (!format_encode_string (&buffer, &sep, "module-path",
+		                           (const unsigned char*)uri->module_path,
+		                           strlen (uri->module_path), 0)) {
+			return_val_if_reached (P11_KIT_URI_UNEXPECTED);
+		}
+	}
+
+	for (i = 0; i < uri->qattrs->num; i++) {
+		Attribute *attr = uri->qattrs->elem[i];
+		if (!format_encode_string (&buffer, &sep, attr->name,
+					   (const unsigned char *) attr->value,
+					   strlen (attr->value), 0)) {
+			return_val_if_reached (P11_KIT_URI_UNEXPECTED);
+		}
+	}
+
 	return_val_if_fail (p11_buffer_ok (&buffer), P11_KIT_URI_UNEXPECTED);
 	*string = p11_buffer_steal (&buffer, NULL);
 	return P11_KIT_URI_OK;
@@ -1292,35 +1510,94 @@ parse_module_info (const char *name_start, const char *name_end,
 }
 
 static int
-parse_extra_info (const char *name_start, const char *name_end,
-		  const char *start, const char *end,
-		  P11KitUri *uri)
+parse_pin_query (const char *name_start, const char *name_end,
+		 const char *start, const char *end,
+		 P11KitUri *uri)
 {
-	unsigned char *pin_source;
+	unsigned char *value;
 
 	assert (name_start <= name_end);
 	assert (start <= end);
 
 	if (str_range_equal ("pinfile", name_start, name_end) ||
 	    str_range_equal ("pin-source", name_start, name_end)) {
-		pin_source = p11_url_decode (start, end, P11_URL_WHITESPACE, NULL);
-		if (pin_source == NULL)
+		value = p11_url_decode (start, end, P11_URL_WHITESPACE, NULL);
+		if (value == NULL)
 			return P11_KIT_URI_BAD_ENCODING;
 		free (uri->pin_source);
-		uri->pin_source = (char*)pin_source;
+		uri->pin_source = (char*)value;
 		return 1;
 	} else if (str_range_equal ("pin-value", name_start, name_end)) {
-		pin_source = p11_url_decode (start, end, P11_URL_WHITESPACE, NULL);
-		if (pin_source == NULL)
+		value = p11_url_decode (start, end, P11_URL_WHITESPACE, NULL);
+		if (value == NULL)
 			return P11_KIT_URI_BAD_ENCODING;
 		free (uri->pin_value);
-		uri->pin_value = (char*)pin_source;
+		uri->pin_value = (char*)value;
 		return 1;
 	}
 
 	return 0;
 }
 
+static int
+parse_module_query (const char *name_start, const char *name_end,
+		    const char *start, const char *end,
+		    P11KitUri *uri)
+{
+	unsigned char *value;
+
+	assert (name_start <= name_end);
+	assert (start <= end);
+
+	if (str_range_equal ("module-name", name_start, name_end)) {
+		value = p11_url_decode (start, end, P11_URL_WHITESPACE, NULL);
+		if (value == NULL)
+			return P11_KIT_URI_BAD_ENCODING;
+		free (uri->module_name);
+		uri->module_name = (char*)value;
+		return 1;
+	} else if (str_range_equal ("module-path", name_start, name_end)) {
+		value = p11_url_decode (start, end, P11_URL_WHITESPACE, NULL);
+		if (value == NULL)
+			return P11_KIT_URI_BAD_ENCODING;
+		free (uri->module_path);
+		uri->module_path = (char*)value;
+		return 1;
+	}
+
+	return 0;
+}
+
+static int
+parse_vendor_query (const char *name_start, const char *name_end,
+		    const char *start, const char *end,
+		    P11KitUri *uri)
+{
+	unsigned char *name;
+	unsigned char *value;
+
+	assert (name_start <= name_end);
+	assert (start <= end);
+
+	/* FIXME: Should we limit the characters in NAME, according to
+	 * the specification?  */
+	name = malloc (name_end - name_start + 1);
+	if (name == NULL)
+		return P11_KIT_URI_BAD_ENCODING;
+	memcpy (name, name_start, name_end - name_start);
+	name[name_end - name_start] = '\0';
+
+	value = p11_url_decode (start, end, P11_URL_WHITESPACE, NULL);
+	if (value == NULL) {
+		free (name);
+		return P11_KIT_URI_BAD_ENCODING;
+	}
+
+	insert_attribute (uri->qattrs, (char *)name, (char *)value);
+
+	return 0;
+}
+
 /**
  * p11_kit_uri_parse:
  * @string: The string to parse
@@ -1387,23 +1664,25 @@ p11_kit_uri_parse (const char *string, P11KitUriType uri_type,
 	uri->module.libraryVersion.major = (CK_BYTE)-1;
 	uri->module.libraryVersion.minor = (CK_BYTE)-1;
 	uri->unrecognized = 0;
+	uri->slot_id = (CK_SLOT_ID)-1;
 	free (uri->pin_source);
 	uri->pin_source = NULL;
 	free (uri->pin_value);
 	uri->pin_value = NULL;
-	uri->slot_id = (CK_SLOT_ID)-1;
+	free (uri->module_name);
+	uri->module_name = NULL;
+	free (uri->module_path);
+	uri->module_path = NULL;
+	p11_array_clear (uri->qattrs);
 
+	/* Parse the path. */
 	for (;;) {
-		spos = strchr (string, ';');
-		if (spos == NULL) {
-			spos = string + strlen (string);
-			assert (*spos == '\0');
-			if (spos == string)
-				break;
-		}
+		spos = string + strcspn (string, ";?");
+		if (spos == string)
+			break;
 
 		epos = strchr (string, '=');
-		if (epos == NULL || spos == string || epos == string || epos >= spos) {
+		if (epos == NULL || epos == string || epos >= spos) {
 			free (allocated);
 			return P11_KIT_URI_BAD_SYNTAX;
 		}
@@ -1423,8 +1702,10 @@ p11_kit_uri_parse (const char *string, P11KitUriType uri_type,
 			ret = parse_module_info (string, epos, epos + 1, spos, uri);
 		if (ret == 0 && (uri_type & P11_KIT_URI_FOR_MODULE_WITH_VERSION) == P11_KIT_URI_FOR_MODULE_WITH_VERSION)
 			ret = parse_module_version_info (string, epos, epos + 1, spos, uri);
+		/* Accept 'pin-source' and 'pin-value' in path
+		 * attributes for backward compatibility.  */
 		if (ret == 0)
-			ret = parse_extra_info (string, epos, epos + 1, spos, uri);
+			ret = parse_pin_query (string, epos, epos + 1, spos, uri);
 
 		if (ret < 0) {
 			free (allocated);
@@ -1433,9 +1714,44 @@ p11_kit_uri_parse (const char *string, P11KitUriType uri_type,
 		if (ret == 0)
 			uri->unrecognized = true;
 
+		string = spos;
 		if (*spos == '\0')
 			break;
-		string = spos + 1;
+		if (*spos == '?')
+			break;
+		string++;
+	}
+
+	/* Parse the query. */
+	for (;;) {
+		if (*string == '\0')
+			break;
+		string++;
+		spos = strchr (string, '&');
+		if (spos == NULL) {
+			spos = string + strlen (string);
+			assert (*spos == '\0');
+			if (spos == string)
+				break;
+		}
+
+		epos = strchr (string, '=');
+		if (epos == NULL || spos == string || epos == string || epos >= spos) {
+			free (allocated);
+			return P11_KIT_URI_BAD_SYNTAX;
+		}
+
+		ret = parse_pin_query (string, epos, epos + 1, spos, uri);
+		if (ret == 0)
+			ret = parse_module_query (string, epos, epos + 1, spos, uri);
+		if (ret == 0)
+			ret = parse_vendor_query (string, epos, epos + 1, spos, uri);
+		if (ret < 0) {
+			free (allocated);
+			return ret;
+		}
+
+		string = spos;
 	}
 
 	free (allocated);
@@ -1457,6 +1773,9 @@ p11_kit_uri_free (P11KitUri *uri)
 	p11_attrs_free (uri->attrs);
 	free (uri->pin_source);
 	free (uri->pin_value);
+	free (uri->module_name);
+	free (uri->module_path);
+	p11_array_free (uri->qattrs);
 	free (uri);
 }
 
