@@ -792,6 +792,7 @@ const static builder_schema certificate_schema = {
 	  { CKA_CERTIFICATE_TYPE, REQUIRE | CREATE, type_ulong },
 	  { CKA_TRUSTED, CREATE | WANT, type_bool },
 	  { CKA_X_DISTRUSTED, CREATE | WANT, type_bool },
+	  { CKA_NSS_MOZILLA_CA_POLICY, CREATE | WANT, type_bool },
 	  { CKA_CERTIFICATE_CATEGORY, CREATE | WANT, type_ulong },
 	  { CKA_CHECK_VALUE, CREATE | WANT, },
 	  { CKA_START_DATE, CREATE | MODIFY | WANT, type_date },
@@ -1288,7 +1289,7 @@ build_trust_object_eku (CK_ATTRIBUTE *object,
 	else
 		neutral = allow;
 
-	/* The value set if a purpose is explictly rejected */
+	/* The value set if a purpose is explicitly rejected */
 	disallow = CKT_NSS_NOT_TRUSTED;
 
 	for (i = 0; eku_attribute_map[i].type != CKA_INVALID; i++) {
