@@ -37,6 +37,7 @@
 #define P11_DEBUG_FLAG P11_DEBUG_TOOL
 
 #include "attrs.h"
+#include "constants.h"
 #include "debug.h"
 #include "oid.h"
 #include "dict.h"
@@ -51,10 +52,10 @@
 #include <string.h>
 
 static bool
-load_stapled_extension (p11_dict *stapled,
-                        p11_dict *asn1_defs,
-                        const unsigned char *der,
-                        size_t len)
+load_attached_extension (p11_dict *attached,
+                         p11_dict *asn1_defs,
+                         const unsigned char *der,
+                         size_t len)
 {
 	char message[ASN1_MAX_ERROR_DESCRIPTION_SIZE];
 	node_asn *ext;
@@ -66,7 +67,7 @@ load_stapled_extension (p11_dict *stapled,
 
 	ext = p11_asn1_decode (asn1_defs, "PKIX1.Extension", der, len, message);
 	if (ext == NULL) {
-		p11_message ("couldn't parse stapled certificate extension: %s", message);
+		p11_message ("couldn't parse attached certificate extension: %s", message);
 		return false;
 	}
 
@@ -83,21 +84,21 @@ load_stapled_extension (p11_dict *stapled,
 	oid = memdup (der + start, length);
 	return_val_if_fail (oid != NULL, false);
 
-	if (!p11_dict_set (stapled, oid, ext))
+	if (!p11_dict_set (attached, oid, ext))
 		return_val_if_reached (false);
 
 	return true;
 }
 
 static p11_dict *
-load_stapled_extensions (p11_enumerate *ex,
-                         CK_ATTRIBUTE *spki)
+load_attached_extensions (p11_enumerate *ex,
+                          CK_ATTRIBUTE *spki)
 {
 	CK_OBJECT_CLASS extension = CKO_X_CERTIFICATE_EXTENSION;
 	CK_ATTRIBUTE *attrs;
 	P11KitIter *iter;
 	CK_RV rv = CKR_OK;
-	p11_dict *stapled;
+	p11_dict *attached;
 
 	CK_ATTRIBUTE match[] = {
 		{ CKA_CLASS, &extension, sizeof (extension) },
@@ -108,12 +109,12 @@ load_stapled_extensions (p11_enumerate *ex,
 		{ CKA_VALUE, },
 	};
 
-	stapled = p11_dict_new (p11_oid_hash, p11_oid_equal,
+	attached = p11_dict_new (p11_oid_hash, p11_oid_equal,
 	                        free, p11_asn1_free);
 
 	/* No ID to use, just short circuit */
 	if (!spki->pValue || !spki->ulValueLen)
-		return stapled;
+		return attached;
 
 	iter = p11_kit_iter_new (NULL, 0);
 	p11_kit_iter_add_filter (iter, match, 2);
@@ -126,9 +127,9 @@ load_stapled_extensions (p11_enumerate *ex,
 			attrs = p11_attrs_buildn (NULL, template, 1);
 			rv = p11_kit_iter_load_attributes (iter, attrs, 1);
 			if (rv == CKR_OK) {
-				if (!load_stapled_extension (stapled, ex->asn1_defs,
-				                             attrs[0].pValue,
-				                             attrs[0].ulValueLen)) {
+				if (!load_attached_extension (attached, ex->asn1_defs,
+				                              attrs[0].pValue,
+				                              attrs[0].ulValueLen)) {
 					rv = CKR_GENERAL_ERROR;
 				}
 			}
@@ -137,13 +138,13 @@ load_stapled_extensions (p11_enumerate *ex,
 	}
 
 	if (rv != CKR_OK && rv != CKR_CANCEL) {
-		p11_message ("couldn't load stapled extensions for certificate: %s", p11_kit_strerror (rv));
-		p11_dict_free (stapled);
-		stapled = NULL;
+		p11_message ("couldn't load attached extensions for certificate: %s", p11_kit_strerror (rv));
+		p11_dict_free (attached);
+		attached = NULL;
 	}
 
 	p11_kit_iter_free (iter);
-	return stapled;
+	return attached;
 }
 
 static bool
@@ -153,8 +154,8 @@ extract_purposes (p11_enumerate *ex)
 	unsigned char *value = NULL;
 	size_t length;
 
-	if (ex->stapled) {
-		ext = p11_dict_get (ex->stapled, P11_OID_EXTENDED_KEY_USAGE);
+	if (ex->attached) {
+		ext = p11_dict_get (ex->attached, P11_OID_EXTENDED_KEY_USAGE);
 		if (ext != NULL) {
 			value = p11_asn1_read (ext, "extnValue", &length);
 			return_val_if_fail (value != NULL, false);
@@ -268,29 +269,32 @@ extract_certificate (p11_enumerate *ex)
 	return true;
 }
 
+static CK_ATTRIBUTE *
+prepare_attr_types (void)
+{
+	CK_ATTRIBUTE *attrs;
+	int i, count;
+
+	/* Count the number of attributes we know about */
+	for (count = 0; p11_constant_types[count].value != CKA_INVALID; count++);
+
+	attrs = calloc (count + 1, sizeof (CK_ATTRIBUTE));
+	return_val_if_fail (attrs != NULL, NULL);
+
+	for (i = 0; i < count; i++)
+		attrs[i].type = p11_constant_types[i].value;
+	attrs[count].type = CKA_INVALID;
+
+	return attrs;
+}
+
 static bool
 extract_info (p11_enumerate *ex)
 {
 	CK_ATTRIBUTE *attr;
 	CK_RV rv;
 
-	static const CK_ATTRIBUTE attr_types[] = {
-		{ CKA_ID, },
-		{ CKA_CLASS, },
-		{ CKA_CERTIFICATE_TYPE, },
-		{ CKA_LABEL, },
-		{ CKA_VALUE, },
-		{ CKA_SUBJECT, },
-		{ CKA_ISSUER, },
-		{ CKA_SERIAL_NUMBER, },
-		{ CKA_TRUSTED, },
-		{ CKA_CERTIFICATE_CATEGORY },
-		{ CKA_X_DISTRUSTED },
-		{ CKA_PUBLIC_KEY_INFO },
-		{ CKA_INVALID, },
-	};
-
-	ex->attrs = p11_attrs_dup (attr_types);
+	ex->attrs = prepare_attr_types ();
 	rv = p11_kit_iter_load_attributes (ex->iter, ex->attrs, p11_attrs_count (ex->attrs));
 
 	/* The attributes couldn't be loaded */
@@ -304,24 +308,26 @@ extract_info (p11_enumerate *ex)
 		return false;
 
 	/* If a certificate then  */
-	if (ex->klass != CKO_CERTIFICATE) {
-		p11_message ("skipping non-certificate object");
-		return false;
-	}
+	if (ex->flags & P11_ENUMERATE_CORRELATE) {
+		if (ex->klass != CKO_CERTIFICATE) {
+			p11_message ("skipping non-certificate object");
+			return false;
+		}
 
-	if (!extract_certificate (ex))
-		return false;
+		if (!extract_certificate (ex))
+			return false;
 
-	attr = p11_attrs_find_valid (ex->attrs, CKA_PUBLIC_KEY_INFO);
-	if (attr) {
-		ex->stapled = load_stapled_extensions (ex, attr);
-		if (!ex->stapled)
+		attr = p11_attrs_find_valid (ex->attrs, CKA_PUBLIC_KEY_INFO);
+		if (attr) {
+			ex->attached = load_attached_extensions (ex, attr);
+			if (!ex->attached)
+				return false;
+		}
+
+		if (!extract_purposes (ex))
 			return false;
 	}
 
-	if (!extract_purposes (ex))
-		return false;
-
 	return true;
 }
 
@@ -337,8 +343,8 @@ extract_clear (p11_enumerate *ex)
 	ex->cert_der = NULL;
 	ex->cert_len = 0;
 
-	p11_dict_free (ex->stapled);
-	ex->stapled = NULL;
+	p11_dict_free (ex->attached);
+	ex->attached = NULL;
 
 	p11_array_free (ex->purposes);
 	ex->purposes = NULL;
@@ -368,6 +374,11 @@ on_iterate_load_filter (p11_kit_iter *iter,
 	if (ex->limit_to_purposes && ex->purposes) {
 		*matches = CK_FALSE;
 		for (i = 0; i < ex->purposes->num; i++) {
+			if (strcmp (ex->purposes->elem[i], P11_OID_ANY_EXTENDED_KEY_USAGE_STR) == 0) {
+				p11_debug ("anyExtendedKeyUsage is set, skipping filtering by purposes");
+				*matches = CK_TRUE;
+				break;
+			}
 			if (p11_dict_get (ex->limit_to_purposes, ex->purposes->elem[i])) {
 				*matches = CK_TRUE;
 				break;
@@ -593,6 +604,9 @@ p11_enumerate_opt_filter (p11_enumerate *ex,
 		attrs = p11_attrs_build (NULL, &certificate, &x509, NULL);
 		ex->flags |= P11_ENUMERATE_COLLAPSE;
 
+	} else if (strcmp (option, "all") == 0) {
+		attrs = p11_attrs_build (NULL, NULL);
+
 	} else {
 		p11_message ("unsupported or unrecognized filter: %s", option);
 		return false;
